// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::ops::Deref;

use serde::{Serialize, Deserialize};

use super::{Bullet, RADIUS};

use crate::player::{Player, forcefield::Forcefield};
use crate::utils::maths::Circle;
use crate::blocks::Blocks;

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
pub struct Manager(Vec<Bullet>);

impl Manager {
	pub fn new() -> Manager { Manager::default() }

	pub fn add(&mut self, bullet: Bullet) {
		self.0.push(bullet);
	}

	pub fn update(&mut self, dt: f32, forcefields: &[Forcefield]) {
		self.0.retain_mut(|bullet| bullet.update(dt, forcefields));
	}

	pub fn disown(&mut self, id: u32) {
		self.0.iter_mut().filter(|b| b.owned_by(id)).for_each(Bullet::disown);
	}

	pub fn collide_player(&mut self, player: &Player, id: u32) -> Vec<Bullet> {
		let mut collided = Vec::new();
		self.0.retain(|bullet| {
			let should_remove = bullet.colliding_circle(player) && !bullet.owned_by(id);
			if should_remove {
				collided.push(bullet.clone()); // Vec::extract_if would be nice but it's only in nightly Rust
			}
			!should_remove
		});
		collided
	}

	pub fn collide_blocks(&mut self, blocks: &Blocks) {
		self.0.retain(|bullet| !blocks.iter_range(bullet.pos, RADIUS).any(|block| (bullet.pos, RADIUS).colliding_rect(&block)));
	}
}

impl Deref for Manager {
	type Target = [Bullet];

	fn deref(&self) -> &Self::Target {
		self.0.as_slice()
	}
}
