// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod lan_discovery;
pub(super) mod router;

use std::{net::SocketAddr, sync::{Arc, RwLock}, thread};

use rustyline::{DefaultEditor, Config, EditMode, error::ReadlineError};
use tokio::{io::{BufWriter, AsyncWriteExt}, fs, runtime::Handle};
use quinn::{Endpoint, Incoming, SendStream, VarInt, ServerConfig, EndpointConfig};
use rustls::pki_types::{CertificateDer, PrivateKeyDer, PrivatePkcs8KeyDer};
use bincode::Options;
use strum::IntoEnumIterator;

use router::{Router, RouterEntry, GameServer};

use super::{CertInfo, SHUTTING_DOWN_MSG, SHUTTING_DOWN_STATUS, repl::{Command, CommandVariant}};

use crate::{ServerArgs, LanDiscoveryConfig};
use crate::net::{serp::{Request, Response, Vu30, vu30::AsyncWriteVu30}, info::Info, message_stream::quic_server::QuicServerMsIncoming, lan_discovery::server::LanDiscoveryServer, message_stream::quic_common};
use crate::utils::net::{WaitIdleTimeout, SocketAddrToCanonical, udp};

pub struct ServerEndpoint {
	endpoint: Endpoint,
	lan_discovery: Option<LanDiscoveryServer>,
	lan_discovery_config: LanDiscoveryConfig,
	router: RwLock<Router>,
	info: Info,
}

impl ServerEndpoint {
	pub async fn build(addr: SocketAddr, cert_info: Option<CertInfo>, router: Router, lan_discovery: Option<LanDiscoveryServer>, lan_discovery_config: LanDiscoveryConfig) -> Result<Arc<ServerEndpoint>, String> {
		// Loads certificate information from the provided paths, otherwise using a self-signed certificate
		let (certs, key) = if let Some(info) = cert_info {
			ServerEndpoint::load_cert_info(info).await.map_err(|err| format!("failed loading certificate info: {err}"))?
		} else {
			let cert = rcgen::generate_simple_self_signed(vec![String::from("localhost")]).map_err(|err| format!("cannot generate self-signed certificate: {err}"))?;
			let key = PrivatePkcs8KeyDer::from(cert.key_pair.serialize_der());
			(vec![cert.cert.into()], key.into())
		};

		let config = ServerConfig::with_single_cert(certs, key).map_err(|err| format!("cannot create server config: {err}"))?;
		let socket = udp::new(addr).map_err(|err| format!("failed binding socket on {addr}: {err}"))?;
		let runtime = quinn::default_runtime().ok_or_else(|| String::from("failed getting tokio runtime"))?;
		let endpoint = Endpoint::new(EndpointConfig::default(), Some(config), socket, runtime).map_err(|err| format!("cannot create QUIC server endpoint on {addr}: {err}"))?;

		Ok(Arc::new(ServerEndpoint { endpoint, lan_discovery, lan_discovery_config, router: RwLock::new(router), info: Info::new() }))
	}

	// Copied and modified from https://github.com/quinn-rs/quinn/blob/main/quinn/examples/server.rs
	async fn load_cert_info(info: CertInfo) -> Result<(Vec<CertificateDer<'static>>, PrivateKeyDer<'static>), String> {
		let cert_data = fs::read(info.cert_path.as_ref()).await.map_err(|err| format!("failed loading certificate chain at \"{}\": {err}", info.cert_path))?;
		let certs = rustls_pemfile::certs(&mut &*cert_data)
			.collect::<Result<_, _>>()
			.map_err(|err| format!("failed parsing certificates, make sure they're in PEM format: {err}"))?;

		let key_data = fs::read(info.key_path.as_ref()).await.map_err(|err| format!("failed loading private key at \"{}\": {err}", info.key_path))?;
		let key = match rustls_pemfile::private_key(&mut &*key_data) {
			Ok(Some(key)) => key,
			Ok(None) => return Err(String::from("no private key found")),
			Err(err) => return Err(format!("failed parsing private key, make sure it's in the PEM format: {err}")),
		};

		Ok((certs, key))
	}

	pub async fn run(self: Arc<Self>, args: ServerArgs, addr: SocketAddr) {
		self.run_lan_discovery();

		let endpoint = Arc::clone(&self);
		let handle = Handle::current();

		/*
		 * Using a regular thread rather than a task because the rustyline crate
		 * isn't async, and I'm not using task::spawn_blocking as that recommends
		 * a thread if the task would last for the entire program.
		 */
		let edit_mode = if args.use_vi_mode { EditMode::Vi } else { EditMode::Emacs };
		thread::spawn(move || endpoint.repl(handle, args, addr, edit_mode));

		while let Some(incoming) = self.endpoint.accept().await {
			let endpoint = Arc::clone(&self);
			tokio::spawn(endpoint.handle_incoming(incoming));
		}

		self.endpoint.wait_idle_timeout().await;
	}

	fn repl(self: Arc<Self>, handle: Handle, args: ServerArgs, addr: SocketAddr, edit_mode: EditMode) {
		let mut rustyline = DefaultEditor::with_config(Config::builder().edit_mode(edit_mode).build()).unwrap();

		loop {
			match rustyline.readline("> ") {
				Ok(line) => {
					match Command::build(&line) {
						Ok(Some(Command::Help(None))) => {
							println!("Showing help of all commands:");
							for cmd in CommandVariant::iter() {
								print!("{}", cmd.help());
							}
						},
						Ok(Some(Command::Help(Some(cmd)))) => print!("{}", cmd.help()),
						Ok(Some(Command::Exit)) => break,
						Ok(Some(Command::Clear)) => {
							if let Err(err) = rustyline.clear_screen() {
								println!("error: clear: failed clearing screen: {err}");
							}
						},
						Ok(Some(Command::Add(entry))) => {
							if let Err(err) = self.add_game_server(&handle, entry, &args, addr) {
								println!("error: add: {err}");
							}
						},
						Ok(Some(Command::Remove(game_id))) => {
							let res = self.router.write().unwrap().remove_game_server(game_id.as_ref()); // Ensures that the lock is dropped as soon as possible
							if let Err(err) = res {
								println!("error: remove: cannot remove game server: {err}");
							}
						},
						Ok(Some(Command::Route(cmd))) => {
							let res = self.router.write().unwrap().change_route(cmd);
							if let Err(err) = res {
								println!("error: route: {err}");
							}
						},
						Ok(Some(Command::Status(id))) => {
							let status = self.router.read().unwrap().get_status(id.as_ref().map(Box::as_ref));
							print!("{status}");
						},
						Ok(Some(Command::Cert(info))) => {
							if let Err(err) = self.update_cert_info(&handle, info) {
								println!("error: cert: {err}");
							} else {
								println!("Successfully updated certificate information!");
							}
						},
						Ok(Some(Command::Info)) => self.info.print(),
						Ok(None) => (),
						Err(err) => println!("error: {err}"), // Not using stderr as that's for logs, and if redirecting logs to a file, those will be ignored
					}

					if let Err(err) = rustyline.add_history_entry(line) {
						println!("error: cannot add to history: {err}");
					}
				},
				Err(ReadlineError::Interrupted) => println!("^C"),
				Err(ReadlineError::Eof) => break,
				Err(ReadlineError::WindowResized) => (), // Don't know how that's an issue
				Err(err) => {
					println!("error reading line, exiting early: {err}");
					break;
				},
			}
		}

		self.endpoint.close(SHUTTING_DOWN_STATUS, SHUTTING_DOWN_MSG);
	}

	fn add_game_server(&self, handle: &Handle, entry: RouterEntry, args: &ServerArgs, addr: SocketAddr) -> Result<(), String> {
		let server = handle.block_on(GameServer::build(addr, &entry, args.sanity_checks, self.lan_discovery.is_some())).map_err(|err| format!("failed creating game server: {err}"))?;
		self.router.write().unwrap().add_game_server(entry, server).map_err(|err| format!("failed adding game server: {err}"))?;
		Ok(())
	}

	fn update_cert_info(&self, handle: &Handle, info: CertInfo) -> Result<(), String> {
		let (certs, key) = handle.block_on(ServerEndpoint::load_cert_info(info)).map_err(|err| format!("failed loading certificate info: {err}"))?;
		let config = ServerConfig::with_single_cert(certs, key).map_err(|err| format!("cannot create server config: {err}"))?;
		self.endpoint.set_server_config(Some(config));
		Ok(())
	}

	async fn handle_incoming(self: Arc<Self>, incoming: Incoming) {
		let addr = incoming.remote_address().to_canonical();
		match incoming.await {
			Ok(connection) => {
				log::info!("connection attempt, addr = {addr}");

				// Doesn't seem to do anything, but maybe that's because my testing had the streams created immediately
				connection.set_max_concurrent_bi_streams(VarInt::from_u32(1));
				connection.set_max_concurrent_uni_streams(VarInt::from_u32(0));

				match Request::receive(&connection).await {
					Ok((Request::Play(game_id), subrequest_index, mut sender, receiver)) => {
						let router_res = self.router.read().unwrap().route(&game_id);
						if let Some((index, stream_sender)) = router_res {
							log::info!("play request received, addr = {addr}, game id = {game_id}, routing to server {index}");

							let Ok(stream) = QuicServerMsIncoming::build(connection, sender, receiver, subrequest_index).await.inspect_err(|err|
								log::info!("creating message stream failed, addr = {addr}, game id = {game_id}, server index = {index}: {err}")
							) else { return; };

							let _ = stream_sender.send(stream).await;
						} else {
							log::info!("connection failed, no server matches game id, addr = {addr}, game id = {game_id}");
							if let Err(err) = Response::Understood(subrequest_index, String::from("no server matches game id")).reply(&mut sender).await {
								log::info!("failed sending understood response, addr = {addr}: {err}");
							}
						}
					},
					Ok((Request::Info, index, sender, _receiver)) => {
						log::info!("info request received, addr = {addr}");
						if let Err(err) = ServerEndpoint::respond_info_request(sender, index, &self.info).await {
							log::info!("sending server info failed, addr = {addr}: {err}");
						}
					},
					Err(Ok((status, mut sender))) => {
						log::info!("connection failed with status, addr = {addr}");
						if let Err(err) = status.reply(&mut sender).await {
							log::info!("sending server info failed, addr = {addr}: {err}");
						}
					},
					Err(Err(err)) => log::info!("connection failed, addr = {addr}: {err}"),
				}
			},
			Err(err) => log::info!("connection attempt failed, addr = {addr}: {err}"),
		}
	}

	async fn respond_info_request(mut sender: BufWriter<SendStream>, index: Vu30, info: &Info) -> Result<(), String> {
		Response::Success(index).reply(&mut sender).await.map_err(|err| format!("failed sending success response: {err}"))?;
		let info_data = quic_common::bincode_options().serialize(info).map_err(|err| format!("failed serialising info: {err}"))?;
		sender.write_bytes(&info_data).await.map_err(|err| format!("failed sending info: {err}"))?;
		sender.flush().await.map_err(|err| format!("failed flushing: {err}"))?;
		let _ = sender.get_mut().stopped().await;
		Ok(())
	}
}
