// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{collections::{BTreeMap, BTreeSet, btree_map::Entry}, net::SocketAddr, num::{NonZeroU16, NonZeroUsize}, sync::{Arc, atomic::{AtomicUsize, Ordering}}, fmt::Write};

use tokio::sync::{mpsc::{self, Sender}, oneshot::Sender as OneshotSender};
use nonempty::NonEmpty;

use crate::net::{server::{self, Server, repl::{RouteCommand, INDENTATION}}, connection_listener::networked::NetworkedConnectionListener, lan_discovery::{LanServerInfo, PROTOCOL_VERSION}};
use crate::net::message_stream::quic_server::QuicServerMsIncoming;
use crate::net::serp::request::{PLAY_VERSION, request_types};

type LanDiscoverySender = Sender<OneshotSender<LanServerInfo>>;

pub struct GameServer {
	stream_sender: Sender<QuicServerMsIncoming>,
	lan_discovery_sender: Option<LanDiscoverySender>,
	path: Arc<str>,
	game_ids: BTreeSet<Arc<str>>,
	player_count: Arc<AtomicUsize>,
	max_players: Option<NonZeroUsize>,
}

impl GameServer {
	pub async fn build(addr: SocketAddr, entry: &RouterEntry, sanity_checks: bool, use_lan_discovery: bool) -> Result<GameServer, String> {
		let config = server::load_config_async(entry.path.as_ref()).await.map_err(|err| format!("failed loading config: {err}"))?;
		let max_players = config.max_players;

		let info = LanServerInfo {
			version: PROTOCOL_VERSION,
			request_type: request_types::PLAY,
			request_version: PLAY_VERSION,
			port: addr.port(),
			cur_players: 0,
			max_players: config.max_players.and_then(|max| u16::try_from(max.get()).ok()).and_then(NonZeroU16::new),
			game_id: Box::from(entry.game_ids.first().as_ref()),
			name: config.lan_discovery.name.clone().into_boxed_str(),
			desc: config.lan_discovery.desc.clone().into_boxed_str(),
		};

		let (sender, receiver) = mpsc::channel(256);

		let connection_listener = NetworkedConnectionListener::new(receiver);
		let server = Server::new(Box::new(connection_listener), config, sanity_checks).map_err(|err| format!("failed creating server: {err}"))?;

		let (lan_info, lan_discovery_sender) = match info.validate() {
			_ if !use_lan_discovery => (None, None),
			Ok(info) => {
				let (sender, receiver) = mpsc::channel(256);
				(Some((info, receiver)), Some(sender))
			},
			Err(err) => {
				log::warn!("invalid LAN discovery config for server at \"{}\": {err}", entry.path);
				(None, None)
			},
		};

		let player_count = Arc::new(AtomicUsize::new(0));
		tokio::spawn(server.run(lan_info, Arc::clone(&player_count)));

		Ok(GameServer {
			stream_sender: sender,
			lan_discovery_sender,
			path: Arc::clone(&entry.path),
			game_ids: entry.game_ids.iter().map(Arc::clone).collect(),
			player_count,
			max_players,
		})
	}
}

pub struct RouterEntry {
	path: Arc<str>,
	game_ids: NonEmpty<Arc<str>>,
}

impl RouterEntry {
	/**
	 * Parses a colon-separated string consisting of "id:id:...:id:path" and
	 * splits up the game ids and paths. If there are no colons, then the server
	 * is treated as the default game server, so it's given the id "default".
	 *
	 * A game server can have the "default" id and multiple others. Those extra
	 * ids don't have any effect on their own, but they can prevent other servers
	 * from being created with those ids, and if the default id is removed, those
	 * ids can still be used.
	 */
	pub fn build(game_server: &str) -> Result<RouterEntry, String> {
		let mut parts: Vec<Arc<str>> = game_server.split(':').map(Arc::from).collect();
		let path = parts.pop().expect("split should return at least one part"); // Shouldn't ever panic

		let game_ids = NonEmpty::from_vec(parts).unwrap_or_else(|| NonEmpty::new(Arc::from("default")));
		let mut set = BTreeSet::new();

		// Don't want any duplicates as they might cause issues later on
		for game_id in &game_ids {
			if !set.insert(game_id.as_ref()) {
				return Err(format!("duplicate game id \"{game_id}\""));
			}
		}

		Ok(RouterEntry { path, game_ids })
	}
}

#[derive(Default)]
pub struct Router { // Routes game ids to servers
	map: BTreeMap<Arc<str>, usize>,
	servers: BTreeMap<usize, GameServer>, // Not using a Vec<GameServer> because I want the index to be constant when servers are removed
	next_index: usize,
}

impl Router {
	/**
	 * Adds a new game server to the router. Keeps it unmodified and in a
	 * consistent state on error.
	 */
	pub fn add_game_server(&mut self, entry: RouterEntry, server: GameServer) -> Result<(), String> {
		// First ensures the game id isn't previously used
		for game_id in &entry.game_ids {
			if self.map.contains_key(game_id) {
				// Returns keeping things unmodified
				return Err(format!("game id \"{game_id}\" used multiple times"));
			}
		}

		while self.servers.contains_key(&self.next_index) {
			self.next_index = self.next_index.wrapping_add(1);
		}

		self.servers.insert(self.next_index, server);
		for game_id in entry.game_ids { self.map.insert(game_id, self.next_index); }

		Ok(())
	}

	pub(super) fn remove_game_server(&mut self, game_id: &str) -> Result<(), String> {
		if let Some(index) = self.map.get(game_id) {
			let server = self.servers.remove(index).unwrap();
			for game_id in &server.game_ids {
				let found = self.map.remove(game_id.as_ref()).is_some();
				debug_assert!(found);
			}
			Ok(())
		} else {
			Err(format!("game id \"{game_id}\" not found"))
		}
	}

	pub(super) fn route(&self, game_id: &str) -> Option<(usize, Sender<QuicServerMsIncoming>)> {
		self.map.get(game_id).or_else(|| self.map.get("default")).map(|index| (*index, self.servers[index].stream_sender.clone()))
	}

	pub(super) fn change_route(&mut self, cmd: RouteCommand) -> Result<(), String> {
		match cmd {
			RouteCommand::Add { id, new_id } => {
				let Some(&index) = self.map.get(id.as_ref()) else {
					return Err(format!("failed adding route: game id \"{id}\" does not exist"));
				};

				if self.map.contains_key(new_id.as_ref()) {
					return Err(format!("failed adding route: new game id \"{new_id}\" already exists"));
				}

				let server = self.servers.get_mut(&index).unwrap();

				let new_id = Arc::from(new_id);
				let not_found = server.game_ids.insert(Arc::clone(&new_id));
				debug_assert!(not_found);

				let not_found = self.map.insert(new_id, index).is_none();
				debug_assert!(not_found);
				Ok(())
			},
			RouteCommand::Remove(id) => {
				let map_entry = match self.map.entry(Arc::from(id)) {
					Entry::Occupied(e) => e,
					Entry::Vacant(e) => return Err(format!("failed removing route: game id \"{}\" does not exist", e.key())),
				};
				let (id, index) = (map_entry.key(), *map_entry.get());

				let server = self.servers.get_mut(&index).unwrap();
				match server.game_ids.len() {
					0 => unreachable!(),
					1 => return Err(format!("failed removing route: game server with id \"{id}\" cannot have route removed, as doing so would remove the server (to remove the server, run `remove {id}`)")),
					_ => (),
				}

				let found = server.game_ids.remove(id);
				debug_assert!(found);

				map_entry.remove();
				Ok(())
			},
		}
	}

	// Returning a string rather than printing as I don't want to perform I/O while a lock's being held
	pub(super) fn get_status(&self, game_id: Option<&str>) -> String {
		let mut status = String::new();
		if let Some(id) = game_id { // Get the status for just that id
			if let Some(&index) = self.map.get(id) {
				let _ = writeln!(&mut status, "Showing status for game server with id \"{id}\":");
				self.write_status_for_index(&mut status, index);
			} else {
				let _ = writeln!(&mut status, "No game server with id \"{id}\" exists.");
			}
		} else if !self.servers.is_empty() {
			let _ = writeln!(&mut status, "Showing status for all game servers:");
			for (&index, server) in &self.servers {
				Router::write_status_for_index_and_server(&mut status, index, server);
				let _ = writeln!(&mut status);
			}
		} else {
			let _ = writeln!(&mut status, "No game servers are running.");
		}

		status
	}

	fn write_status_for_index(&self, status: &mut String, index: usize) {
		Router::write_status_for_index_and_server(status, index, &self.servers[&index]);
	}

	fn write_status_for_index_and_server(status: &mut String, index: usize, server: &GameServer) {
		let _ = write!(status, "{index}: {:?}\n{}path = {}\n{INDENTATION}player count = {}", server.game_ids, INDENTATION, server.path, server.player_count.load(Ordering::Relaxed));
		if let Some(max) = server.max_players {
			let _ = write!(status, "/{max}");
		}
		let _ = writeln!(status);
	}

	pub(super) fn get_lan_discovery_senders(&self) -> Vec<LanDiscoverySender> {
		self.servers.values().filter_map(|server| server.lan_discovery_sender.clone()).collect()
	}
}
