// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{fs, path::Path, rc::Rc};

use glium::{Surface, VertexBuffer, Display, Texture2d, Program, Frame, DrawParameters, index::{NoIndices, PrimitiveType}, uniforms::SamplerWrapFunction, uniform, implement_vertex};
use glutin::surface::WindowSurface;
use glam::{Vec2, Mat4};

use crate::game::filesystem::Filesystem;
use crate::world::smooth::LocalPlayer;
use crate::utils::glium_resize;

pub struct Renderer {
	v_vbo: VertexBuffer<Vertex>,
	i_vbo: VertexBuffer<Instance>,
	instance_buf: Vec<Instance>,
	texture: Rc<Texture2d>,
	program: Program,
}

#[derive(Clone, Copy)]
struct Vertex { v_pos: [f32; 2], v_tex_coords: [f32; 2] }
implement_vertex!(Vertex, v_pos, v_tex_coords);

#[derive(Clone, Copy)]
struct Instance { i_pos: [f32; 2], i_dir: [f32; 2], i_colour: [f32; 3], i_tint: f32 }
implement_vertex!(Instance, i_pos, i_dir, i_colour, i_tint);

impl Renderer {
	pub fn new(display: &Display<WindowSurface>, fs: &Filesystem, texture: Rc<Texture2d>) -> Renderer {
		static VERTICES: [Vertex; 12] = [
			// Top side
			Vertex { v_pos: [-0.5, 0.5], v_tex_coords: [0.0, 0.0] },
			Vertex { v_pos: [0.25, 0.375], v_tex_coords: [0.75, 0.25] },
			Vertex { v_pos: [0.5, 0.0], v_tex_coords: [1.0, 1.0] },

			Vertex { v_pos: [-0.5, 0.5], v_tex_coords: [0.0, 0.0] },
			Vertex { v_pos: [-0.25, 0.0], v_tex_coords: [0.25, 1.0] },
			Vertex { v_pos: [0.5, 0.0], v_tex_coords: [1.0, 1.0] },

			// Bottom side
			Vertex { v_pos: [-0.5, -0.5], v_tex_coords: [0.0, 0.0] },
			Vertex { v_pos: [0.25, -0.375], v_tex_coords: [0.75, 0.25] },
			Vertex { v_pos: [0.5, 0.0], v_tex_coords: [1.0, 1.0] },

			Vertex { v_pos: [-0.5, -0.5], v_tex_coords: [0.0, 0.0] },
			Vertex { v_pos: [-0.25, 0.0], v_tex_coords: [0.25, 1.0] },
			Vertex { v_pos: [0.5, 0.0], v_tex_coords: [1.0, 1.0] },
		];

		let vsh = fs::read_to_string(fs.get(Path::new("shaders/player.vsh"))).unwrap();
		let fsh = fs::read_to_string(fs.get(Path::new("shaders/player.fsh"))).unwrap();

		Renderer {
			v_vbo: VertexBuffer::immutable(display, &VERTICES).unwrap(),
			i_vbo: VertexBuffer::empty_persistent(display, 1).unwrap(),
			instance_buf: Vec::with_capacity(1),
			texture,
			program: Program::from_source(display, &vsh, &fsh, None).unwrap(),
		}
	}

	pub fn get_colour(id: u32) -> [f32; 3] {
		match id % 8 {
			0 => [1.0, 0.25, 0.125],
			1 => [0.125, 0.25, 1.0],
			2 => [0.25, 1.0, 0.125],
			3 => [1.0, 1.0, 0.125],
			4 => [0.125, 1.0, 1.0],
			5 => [1.0, 0.125, 1.0],
			6 => [1.0, 1.0, 1.0],
			_ => [0.5, 0.5, 0.5],
		}
	}

	pub fn push(&mut self, player: &LocalPlayer, id: u32, tint: f32, time_diff: f32) {
		self.instance_buf.push(Instance {
			i_pos: player.get_pos(time_diff).into(),
			i_dir: Vec2::from(player.get_dir()).into(),
			i_colour: Renderer::get_colour(id),
			i_tint: tint,
		});
	}

	pub fn render(&mut self, display: &Display<WindowSurface>, frame: &mut Frame, params: &DrawParameters, proj_matrix: &Mat4) {
		let count = self.instance_buf.len();
		glium_resize::vbo_persistent(&mut self.i_vbo, display, count);

		{
			let mut buf = self.i_vbo.map_write();
			for (i, inst) in self.instance_buf.drain(..).enumerate() {
				buf.set(i, inst);
			}
		}

		let uniforms = uniform! { u_matrix: proj_matrix.to_cols_array_2d(), u_texture: self.texture.sampled().wrap_function(SamplerWrapFunction::Clamp) };
		frame.draw((&self.v_vbo, self.i_vbo.slice(..count).unwrap().per_instance().unwrap()), NoIndices(PrimitiveType::TrianglesList), &self.program, &uniforms, params).unwrap();
	}
}
