// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod decorations;
mod hud;

use std::{rc::Rc, cell::RefCell, path::Path};

use glium::{Frame, Surface, DrawParameters, Blend, BlendingFunction};
use winit::{window::Window, event::{WindowEvent, MouseScrollDelta}, dpi::PhysicalSize};
use glam::UVec2;
use png::ColorType;

use decorations::Decorations;
use hud::Hud;

use super::TimeArgs;
use super::model::Model;
use super::controller::Controller;

use crate::game::{Game, config::{Config, action::ActionEvent}};
use crate::world::{SmoothWorld, WorldRenderer, changes::Hit as ChangesHit};
use crate::camera::Camera;
use crate::utils::blending::{ALPHA_BLENDING, ADDITIVE_BLENDING};
use crate::texture;

pub(super) struct View<'a> {
	pub smooth_world: SmoothWorld,
	pub camera: Camera,
	params: DrawParameters<'a>,
	decorations: Decorations,
	world_renderer: WorldRenderer,
	hud: Hud,
}

impl<'a> View<'a> {
	pub fn new(config: Rc<RefCell<Config>>, game: &Game) -> View<'a> {
		let smooth_world = SmoothWorld::new();
		let winsize = { let PhysicalSize { width, height } = game.window.inner_size(); UVec2::new(width, height) };
		let camera = Camera::new(winsize);
		let player_texture = Rc::new(texture::load(&game.display, &game.fs.get(Path::new("textures/player.png")), ColorType::GrayscaleAlpha));
		let world_renderer = WorldRenderer::new(&game.display, &game.fs, Rc::clone(&player_texture));
		let decorations = Decorations::new(config, &game.display, &game.fs, player_texture);

		View {
			smooth_world,
			camera,
			params: DrawParameters {
				blend: Blend {
					color: ALPHA_BLENDING,
					alpha: BlendingFunction::Max, // Differs from Blend::alpha_blending, makes sure that the alpha channel is always 1
					constant_value: (0.0, 0.0, 0.0, 0.0),
				},
				.. DrawParameters::default()
			},
			decorations,
			world_renderer,
			hud: Hud::new(&game.display, &game.fs, winsize, game.window.scale_factor() as f32),
		}
	}

	pub fn reset(&mut self, window: &Window, model: &Model) {
		let size = {
			let size = window.inner_size();
			UVec2::new(size.width, size.height)
		};

		self.hud.reset(size, window.scale_factor() as f32);
		self.smooth_world.reset(model);
		self.camera.reset(self.smooth_world.get_player_pos(0.0), self.smooth_world.get_player_vel(), size);
		self.decorations.reset(&self.camera);
		self.world_renderer.reset(model);
	}

	pub fn action(&mut self, action: ActionEvent) {
		self.camera.action(action);
	}

	pub fn window_event(&mut self, event: &WindowEvent) {
		match event {
			WindowEvent::Resized(size) => {
				let winsize = UVec2::new(size.width, size.height);
				self.camera.resize_event(winsize);
				self.hud.resize_event(winsize);
			}
			WindowEvent::ScaleFactorChanged { scale_factor, .. } => self.hud.scale_factor_event(*scale_factor as f32),
			WindowEvent::MouseWheel { delta: MouseScrollDelta::LineDelta(_, dy), .. } => self.camera.mouse_scroll_event(*dy),
			WindowEvent::MouseWheel { delta: MouseScrollDelta::PixelDelta(pos), .. } => self.camera.mouse_scroll_event(pos.y as f32 * 0.0625),
			_ => (),
		}
	}

	pub fn on_hit(&mut self, hit: &ChangesHit) { self.decorations.on_hit(hit); }

	pub fn render(&mut self, game: &mut Game, frame: &mut Frame, model: &Model, controller: &mut Controller, time: &TimeArgs) {
		/*
		 * Need to subtract dt from the time difference to make the camera move
		 * smoothly when the frame rate varies.
		 *
		 * Not too sure why this is needed, but one reason I suspect is because
		 * the calculation involves the velocity for calculating how the camera's
		 * position should exponentially decay towards the player, using the
		 * previous player's movement from last frame to this frame.
		 *
		 */
		let moved_by = self.camera.update(self.smooth_world.get_player_pos(time.time_diff - time.dt), self.smooth_world.get_player_vel(), time.dt);

		let matrix = self.camera.get_matrix();

		// Updates the player's cursor position with the new player position, to ensure the player is always rendered as looking towards the cursor
		controller.pcontrol.mouse_move_event(self.camera.pixel_to_world_coords(controller.cursor_pos) - self.smooth_world.get_player_pos(time.time_diff));
		self.smooth_world.control_player(&controller.pcontrol);

		// First renders the background, stars and border
		frame.clear_color_srgb(0.0, 0.0, 0.0, 1.0);
		self.decorations.render_stars(time.dt, &game.display, &self.camera, -moved_by, frame, &mut self.params);
		self.world_renderer.render_border(frame, &self.params, matrix);

		// Second renders the particles (including the death effect particles), they are decorative so they can go below everything else
		// Note that updating must come BEFORE adding new particles to avoid any new thrust particles getting a "head start" and being off-centre
		self.decorations.update_particles(time.dt);
		self.decorations.add_particles(time.dt, time.time_diff, &self.smooth_world, model.world.get_powerups());
		self.decorations.render_particles(time.dt, &game.display, frame, &mut self.params, matrix);

		// Third renders everything in the world
		self.world_renderer.render(time.dt, &game.display, frame, &mut self.params, time.time_diff, &model.world, &mut self.smooth_world, &model.blocks, &self.camera);

		// Fourth renders the health bars, ammo status and scores, which are above everything else
		self.hud.push_health_bars(&self.smooth_world, &self.camera, time.time_diff);
		self.hud.render(game, frame, &mut self.params, &model.world, model.id, model.client.get_scores());
	}
}
