// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod action;

use std::{any::Any, rc::Rc, cell::RefCell};

use egui::{RichText, Color32, Button, Margin, Frame, Ui, Window, TextFormat, FontId, Slider, CollapsingHeader, text::LayoutJob};
use strum::IntoEnumIterator;
use glium::Frame as GliumFrame;
use winit::event::Event;

use action::Category;

use crate::game::{Game, config::{Config, DEFAULT_STAR_DENSITY, action::{Action, ActionEvent, ActionState, ActionManager, Input}}, state::{GameState, Status}};
use crate::gui::{GuiState, LABEL_SIZE, style::{UiExt, FontSize, H3_SIZE, STYLE_SETTINGS_BIND_BUTTON}};
use crate::utils::ToStr;

struct Section {
	header: &'static str,
	actions: Vec<Action>,
}

pub struct Settings {
	config: Rc<RefCell<Config>>,
	gui_state: GuiState,
	changed: bool,
	sections: Box<[Section]>,
	opened_window: Option<Action>,
	selecting: bool,
	selected_input: Option<Input>,
	selected_input_string: String,
}

const SELECTING_OPACITY: f32 = 0.375;

impl Settings {
	pub fn new(config: Rc<RefCell<Config>>) -> Settings {
		let mut sections: Box<[Section]> = Category::iter()
			.map(|category| Section {
				header: category.to_str(),
				actions: Vec::new(),
			}).collect();

		for action in Action::iter() {
			sections[Category::from(action) as usize].actions.push(action);
		}

		Settings {
			config,
			gui_state: GuiState::new(),
			changed: false,
			sections,
			opened_window: None,
			selecting: false,
			selected_input: None,
			selected_input_string: String::new(),
		}
	}

	fn add_input_list(action_manager: &ActionManager, action: Action, string: &mut String, separator: &str) {
		if let Some(inputs) = action_manager.inputs_of_action(action) {
			for (i, input) in inputs.enumerate() {
				if i != 0 {
					string.push_str(separator);
				}
				string.push_str(&input.to_string());
			}
		}
	}

	#[must_use]
	fn bind_button_clicked(ui: &mut Ui, text: &str, enabled: bool) -> bool {
		ui.add_enabled(enabled, Button::new(RichText::new(text).font_size(STYLE_SETTINGS_BIND_BUTTON))).clicked()
	}
}

impl GameState for Settings {
	fn enable(&mut self, game: &mut Game) {
		self.gui_state.enable(&mut game.gui);
		debug_assert!(!self.selecting);
	}

	fn push(&mut self, _game: &mut Game, msg: Option<Box<dyn Any>>) {
		debug_assert!(msg.is_none());
	}

	fn disable(&mut self, game: &mut Game) {
		if self.changed {
			if let Err(err) = self.config.borrow().save(&game.fs) {
				log::warn!("cannot save config: {err}");
			}

			self.changed = false;
		}

		debug_assert!(!self.selecting);
	}

	fn action(&mut self, action: ActionEvent) {
		self.gui_state.action(action);
	}

	fn event(&mut self, game: &mut Game, event: Event<()>) {
		if let Event::WindowEvent { event, .. } = &event {
			if self.selecting {
				if let Some((input, ActionState::Pressed)) = Input::from_event(event) {
					self.selecting = false;
					self.selected_input = Some(input);
					self.selected_input_string = input.to_string();
					self.config.borrow_mut().action_manager.set_enabled(true);
					return;
				}
			}
		}

		self.gui_state.event(game, &event);
	}

	fn loop_iter(&mut self, game: &mut Game, frame: &mut GliumFrame) -> Status {
		let config = &mut self.config.borrow_mut();

		let mut change: Option<(Action, bool)> = None;
		let mut back_button_clicked = false;

		let decrease_opacity = self.selecting;

		GuiState::update(&mut game.gui, "settings", &game.window, |ctx, ui| {
			if decrease_opacity {
				ui.set_opacity(SELECTING_OPACITY);
			}

			ui.h1("Settings").add();
			ui.h2("Input").add();

			Frame::none().outer_margin(Margin::symmetric(80.0, 0.0)).show(ui, |ui| {
				for section in &self.sections {
					CollapsingHeader::new(RichText::new(section.header).font_size(H3_SIZE)).show(ui, |ui| {
						ui.horizontal_wrapped(|ui| {
							for &action in &section.actions {
								let response = ui.add(Button::new(RichText::new(action.to_str()).font_size(H3_SIZE)));

								if response.clicked() {
									self.opened_window = Some(action);
								}

								response.on_hover_ui_at_pointer(|ui| {
									ui.label(RichText::new("Bound to:").strong());

									let mut inputs = String::new();
									Settings::add_input_list(&config.action_manager, action, &mut inputs, "\n");
									ui.label(inputs);
								});
							}
						});

						ui.add_space(10.0);
					});
				}

				ui.h2("Graphics").add();

				let mut show_stars = config.graphics.star_density > 0.0;
				let prev_show_stars = show_stars;
				ui.vertical(|ui| ui.checkbox(&mut show_stars, "Stars"));

				if prev_show_stars && !show_stars {
					config.graphics.star_density = 0.0;
					self.changed = true;
				} else if !prev_show_stars && show_stars {
					config.graphics.star_density = DEFAULT_STAR_DENSITY;
					self.changed = true;
				}

				if ui.add_enabled(show_stars, Slider::new(&mut config.graphics.star_density, 1.0..=10000.0).text(RichText::new("Star Density (per Megapixel)").color(Color32::from_gray(0xbf))).logarithmic(true)).changed() {
					self.changed = true;
				}

				ui.h3("Particles").add();

				// Don't want vertical centred to be consistent with the other buttons
				ui.vertical(|ui| {
					ui.checkbox(&mut config.graphics.particles.thrust, "Thrust");
					ui.checkbox(&mut config.graphics.particles.powerup, "Powerup");
					ui.checkbox(&mut config.graphics.particles.death, "Death");
					ui.checkbox(&mut config.graphics.particles.hit, "Hit");
				});
			});

			back_button_clicked |= ui.b_back().clicked();

			let (text_colour, strong_colour) = {
				let style = ctx.style();
				(style.visuals.text_color(), style.visuals.strong_text_color())
			};

			if let Some(action) = self.opened_window {
				let mut open = true;
				Window::new(format!("Change Bindings for {}", action.to_str()))
					.open(&mut open)
					.resizable(false)
					.show(ctx, |ui| {
						if decrease_opacity {
							ui.set_opacity(SELECTING_OPACITY);
						}

						/*
						 * I had lots of fun writing this code just to get a simple
						 * string with one part bold and another part non-bold.
						 *
						 * At least I had the documentation...
						 */
						let mut job = LayoutJob::default();
						job.append("Bindings: ", 0.0, TextFormat {
							font_id: FontId::proportional(LABEL_SIZE),
							color: strong_colour,
							.. Default::default()
						});

						let mut inputs = String::new();
						Settings::add_input_list(&config.action_manager, action, &mut inputs, ", ");

						job.append(&inputs, 0.0, TextFormat {
							font_id: FontId::proportional(LABEL_SIZE),
							color: text_colour,
							.. Default::default()
						});

						ui.label(job);
						ui.add_space(10.0);

						if Settings::bind_button_clicked(ui, &format!("Select {}", self.selected_input_string), !decrease_opacity) {
							self.selecting = true;
							self.selected_input_string = String::from("(…)");
							config.action_manager.set_enabled(false);
						}

						ui.horizontal(|ui| {
							let input_bound = self.selected_input.map(|input| config.action_manager.is_bound_to(input, action));
							if Settings::bind_button_clicked(ui, "Bind", input_bound == Some(false) && !decrease_opacity) {
								change = Some((action, true));
							}

							if Settings::bind_button_clicked(ui, "Unbind", input_bound == Some(true) && !decrease_opacity) {
								change = Some((action, false));
							}
						});
					});

				if !open {
					self.opened_window = None;
				}
			}
		});

		if let Some((action, bind)) = change {
			let func = if bind { ActionManager::bind } else { ActionManager::unbind };
			func(&mut config.action_manager, self.selected_input.unwrap(), action);
			self.changed = true;
		}

		game.gui.render(&game.display, frame);

		if back_button_clicked {
			Status::PopState
		} else if self.gui_state.should_exit() {
			if self.opened_window.is_some() {
				self.opened_window = None;
				Status::Ok
			} else {
				Status::PopState
			}
		} else { Status::Ok }
	}
}
